/****************************************************************************

    Neural Network Library
    Copyright (C) 1999 Filin Alexey

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.
              
 ****************************************************************************/

// MultiLayer Perceptron (MLP) simulator (class)

#ifndef _MULTI_LAYER_PERCEPTRON_H_
#  define _MULTI_LAYER_PERCEPTRON_H_

#include <math.h>
#include <cstring>
#include <iostream.h>
#include <assert.h>

#include "mozg/types.hh"
#include "mozg/MLPOutFunc.hh"
#include "mozg/VectorMatrix.hh"


class MLPerceptron

{

public:

// PLEASE DEFINE THE FUNCTION:
// The function is called by learnNet b),c) which send it overall number of
// epoch since learning begin, learning error and test error if it is calcu-
// lated; if test error isn't calculated (for learnNet b)) zero is put in
// second signature parameter of the function.
// The function MUST RETURN true if network learning must be ended (error is
// small enough) then return from learnNet() happens and false in the other
// case.
  bool answerMessage(mozgflt,    // learning error
		     mozgflt);   // test error

// network layer
  struct struct_layer
  {
    mozgVector<mozgflt,mozgint>* outputs,  // neuron outputs
                               * nets,     // weighted inputs sum for non-RBF
                               * errors,   // layer errors
                               * sigmas;   // sigma - Radial Basis Function

    mozgMatrix<mozgflt,mozgint>* weights,     // neuron weights
                               * last_deltas, // neuron weights last deltas
                               * RBFweights;  // need for the backpropagate in
                                        //  the case of the RBF output function

    mozgint num_outputs,      // number of units in layer
            function_number;  // number of output function in layer

    OUT_FUNC lay_output_func; // output functions in layer

    OUT_FUNC_DERIV lay_output_func_deriv; // its derivative

    mozgflt eta,         // learning rate in layer
            temperature; // inverse temperature in layer's output function

  };

  struct_layer* layer;           // pointer to the massive of network layers

  mozgVector<mozgflt,mozgint>* diff,   // difference between desiring and real
                                       // output vectors, NOT FOR Cross-Entropy
                                       // and log-square error
                             * temp_vector; // for Cross-Entropy function

  mozgflt alpha,                   // momentum term coefficient
          sigma_Langevin,          // Langevin Gauss begin sigma
          sigma_Langevin_decrease, // Langevin Gauss sigma coef decrease
          decay,                   // weight decay coefficient
          RBF_sigma_min,           // alowed RBF sigma min
          rate_coeff,              // coeff of changing of rate (D=1.)
          temperature_coeff;       // coeff of changing of temperature (D=1.)

  mozgint num_layers,          // the number of layers (including input layer)
          switch_of_energy_function, // network's energy function
                                     // (called error function)
          overall_epnum; // Overall number of epoch since learning begin

  mozgflt error_over_outputs,     // square of a output error vector module
          error_threshold;       // learning or test error which can be
                                  // allowed to finish learning

  bool flag_of_bias_term, // flag of bias term
       is_call_aM;        // flag of answerMessage() call

// create network
  MLPerceptron (mozgint,        //  number of layers (including input layer)
		mozgint*,       //  number of units in layers (massive)
		mozgint*,       //  output function numbers at layers (massive)
		mozgflt,        //  begin weight dispersion
		bool);          //  flag of bias term

// put learning parameters for "vanilla" backpropagation rule
  void putLParams (mozgflt learning_rate,     // common learning rate
                   mozgflt temperature,       // temperature in out. function
		   mozgflt RBF_sigma_minimum, // alowed RBF sigma min
		   mozgint energy_function)   // switch of energy function
  {
    for (mozgint i=1; i<num_layers; i++) {

      layer[i].eta = learning_rate/layer[i].num_outputs;

      layer[i].temperature = temperature;
    }
    if (((energy_function==2)||(energy_function==4))
	&&(!flag_of_temp_vector)) {

      temp_vector =
	new mozgVector<mozgflt,mozgint> (layer[num_layers-1].num_outputs);

      flag_of_temp_vector = true;
    }
    RBF_sigma_min             = RBF_sigma_minimum;
    switch_of_energy_function = energy_function;
    alpha                     = 0.;
    sigma_Langevin            = 0.;
    sigma_Langevin_decrease   = 0.;
    decay                     = 0.;
    flag_of_optimisation      = true;
  }

// put learning parameters for backpropagation with extensions rule
  void putLParams (mozgflt learning_rate,     // common learning rate
                   mozgflt temperature,       // temperature in out. function
		   mozgflt RBF_sigma_minimum, // alowed RBF sigma min
		   mozgint energy_function,   // switch of energy function
		   mozgflt mom_term_coef,     // alpha - momentum term
		                              // coefficient
		   mozgflt lang_noise_begin_dispersion, // begin Langevin Gauss
		                                        // dispersion
		   mozgflt lang_noise_disp_decr_coef,// Langevin Gauss dis-
                                                     // persion decreasing coef
		   mozgflt weight_decay_coef)     // weight decay coefficient
  {
    for (mozgint i=1; i<num_layers; i++) {

      layer[i].eta = learning_rate/layer[i].num_outputs;

      layer[i].temperature = temperature;

      if (mom_term_coef && !flag_of_last_deltas) //memory is allocated one time
	{
	  layer[i].last_deltas =
	    new mozgMatrix<mozgflt,mozgint> (layer[i].num_outputs,
					     layer[i-1].num_outputs);
	}
    }
    if (((energy_function==2)||(energy_function==4))
	&&(!flag_of_temp_vector)) {

      temp_vector =
	new mozgVector<mozgflt,mozgint> (layer[num_layers-1].num_outputs);

      flag_of_temp_vector = true;
    }
    if (mom_term_coef && !flag_of_last_deltas) flag_of_last_deltas = true;

    RBF_sigma_min             = RBF_sigma_minimum;
    switch_of_energy_function = energy_function;
    alpha                     = mom_term_coef;
    sigma_Langevin            = lang_noise_begin_dispersion; 
    sigma_Langevin_decrease   = lang_noise_disp_decr_coef;
    decay                     = weight_decay_coef;
// may be there's possibility of using of o_adjust_lay_weights ()
    flag_of_optimisation      = !(alpha || decay || sigma_Langevin);
  }

// learn network
  void learnNet (mozgflt* linputs,   // learning input vector
		 mozgflt* loutputs); // learning output vector

// learn network
  void learnNet (mozgflt** linputs,  // learning input vector array
		 mozgflt** loutputs, // learning output vector array
		 mozgint   lssize,   // learning set size
		 mozgint   vec_order,// order of vectors used in learning
                 mozgint   epnum,    // number of epochs network must be
                                     // learned
                 mozgint   pepnum);  // answerMessage() is called for
                                     // each pepnum-th epoch

// learn and test network
  void learnNet (mozgflt** linputs,   // learning input vector array
		 mozgflt** loutputs,  // learning output vector array
		 mozgint   lssize,    // learning set size
		 mozgint   vec_order, // order of vectors used in learning
                 mozgint   epnum,     // number of epochs network must be
                                      // learned
                 mozgint   pepnum,    // answerMessage() is called for
                                      // each pepnum-th epoch
		 mozgflt** tinputs,   // test input vector array
		 mozgflt** toutputs,  // test output vector array
		 mozgint   tssize);   // test set size

// test network
  mozgflt testNet (mozgflt** tinp,     // test input vector array
		   mozgflt** tout,     // test output vector array
		   mozgint   tssize);  // test set size

// propagate input vector forward through network to the outputs and save
// outputs in the outvec
  void propInputs (mozgflt* invec,
		   mozgflt* outvec);

// propagate vectors of the invecarr forward through network to the outputs
// and save outputs for each vector in outvecarr
  void propInputs (mozgflt** invecarr,
		   mozgflt** outvecarr,
		   mozgint   ssize);

// delete network
  ~MLPerceptron();

// put net_inputs to the net inputs
  void putInputs (mozgflt* net_inputs)
  {
    mozgflt* inputs = layer[0].outputs->vec ();

    mozgint num_inputs;

    if (flag_of_bias_term)
      {
	num_inputs = layer[0].num_outputs-1;
	inputs[num_inputs] = 1.;
      }
    else num_inputs = layer[0].num_outputs;

    for (mozgint i=0; i < num_inputs; i++)
      *inputs++ = *net_inputs++;
  }

// put last layer outputs to the net_outputs
  void getOutputs (mozgflt* net_outputs)
  {
    mozgflt* outputs = layer[num_layers-1].outputs->vec ();

    for (mozgint i=0; i < layer[num_layers-1].num_outputs; i++)
      *net_outputs++ = *outputs++;
  }
// propagate layer's inputs through the layer
  void propLayer (const struct_layer&,  // layer where signal go from
		  struct_layer*);       // layer where signal go to

// compute errors for last layer
  void compOutputErrors (mozgflt*,  // target output vector
			 bool);     // compute_output_errors flag

// propagate errors from last layer to first one
  void backpropErrors ();

// obviously...
  void adjustWeights ();

protected:

// compute neuron outputs from neuron nets for the current layer
  void compute_lay_outputs (const mozgVector<mozgflt,mozgint>& lay_nets,
			    const OUT_FUNC output_function,
			    const mozgVector<mozgflt,mozgint>& lay_sigmas,
			    mozgVector<mozgflt,mozgint>* lay_outputs,
			    mozgflt lay_temperature);

// compute neuron errors for the output layer
  void compute_output_errors (const mozgVector<mozgflt,mozgint>& lay_errors,
			      const mozgVector<mozgflt,mozgint>& lay_nets,
			      const mozgVector<mozgflt,mozgint>& lay_outputs,
			      const mozgVector<mozgflt,mozgint>& lay_sigmas,
			      const OUT_FUNC_DERIV output_function_deriv,
			      mozgVector<mozgflt,mozgint>* lay_deltas,
			      mozgflt lay_temperature);

// compute neuron errors for the current layer
  void compute_lay_errors (const mozgVector<mozgflt,mozgint>& prev_errors,
			   const mozgMatrix<mozgflt,mozgint>& prev_weights,
			   const mozgVector<mozgflt,mozgint>& cur_nets,
			   const mozgVector<mozgflt,mozgint>& cur_outputs,
			   const mozgVector<mozgflt,mozgint>& cur_sigmas,
			   const OUT_FUNC_DERIV output_function_deriv,
			   mozgVector<mozgflt,mozgint>* cur_errors,
			   mozgflt curent_temperature);

// update weights for current layer
  void adjust_lay_weights (const struct_layer& prev_layer,
			   struct_layer* cur_layer);

// optimised adjust_lay_weights (), without error backprop extentions
  void o_adjust_lay_weights (const struct_layer& prev_layer,
		             struct_layer* cur_layer);

  bool flag_of_optimisation;   // switch between adjust_lay_weights () and 
                               // o_adjust_lay_weights () (see MLPerceptron ())

  bool flag_of_last_deltas;    // to don't allocate the memory for last_deltas
                               // matrices twice. DON'T REMOVE: LEAD TO MEMORY
                               // LEAKAGE FOR N-FOLD CALL OF putLParams[b]

  bool flag_of_temp_vector;    // is rised when temp_vector is allocated

};

// It is overloading of "file_stream << [MLPerceptron object]"
ofstream& operator << (ofstream& output,
		       const MLPerceptron& net);

// It is overloading of "file_stream >> [MLPerceptron object]"
ifstream& operator >> (ifstream& input,
		       MLPerceptron** p_p_net);

// show net on display
ostream& operator << (ostream& output,
		      const MLPerceptron& net);

#endif // _MULTI_LAYER_PERCEPTRON_H_
