/****************************************************************************

    Neural Network Library
    Copyright (C) 1999 Filin Alexey

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.
              
 ****************************************************************************/
//  Kernel library.
//  mozgVector and mozgMatrix classes and operating with them objects:
// adding, out-product, inner-product, sorting (increasing)...

#ifndef _MOZG_VECTOR_MATRIX_H_
#  define _MOZG_VECTOR_MATRIX_H_

#include <math.h>
#include <iostream.h>
#include <assert.h>
#include <fstream.h>
#include <cstring>

#include "mozg/types.hh"
#include "mozg/config.h"
#include "mozg/before_exception.hh"


template <class vec_t, class vec_size_t> class mozgVector
{
private:

  vec_t* vector;           // pointer to first vector element

  vec_size_t vector_size;  // vector size

public:

// return vector size
  vec_size_t size () const
    {
      return vector_size;
    }

// return vector pointer
  vec_t* vec () const
    {
      return vector;
    }

// return reference to vector element
  vec_t& operator [] (const vec_size_t number) const
  {
    return vector[number];
  }

// multiply each element of first vector by relevant element of second one
  mozgVector&operator *= (const mozgVector& vector2);

// divide each vector element by the number
  mozgVector& operator /= (const vec_t number);

// divide each vector element by relevant denominator element
  mozgVector& operator /= (const mozgVector& denominator);

// create a vector
  mozgVector (vec_size_t size)
    {
      vector_size = size;

      vector = new vec_t[size];
      assert (vector);
    }

// create the vector
  mozgVector (vec_size_t size,mozgflt number)
    {
      mozgint stmp = size;

      vector_size = size;

      vector = new vec_t[size];
      assert (vector);

      mozgflt* tmp = vector;
      while (stmp--)
	*tmp++ = number;
    }

// delete vector
  ~mozgVector()
    {
      delete [] vector;
    }
};

// It is overloading "file_stream << [mozgVector object]"
template <class vec_t, class vec_size_t>
ofstream& operator << (ofstream& output,
		       const mozgVector<vec_t,vec_size_t>& vector);

// It is overloading "file_stream >> [mozgVector object]"
template <class vec_t, class vec_size_t>
ifstream& operator >> (ifstream& input,
		       const mozgVector<vec_t,vec_size_t>& vector);

// It shows vector
template <class vec_t, class vec_size_t>
ostream& operator << (ostream& output,
		      const mozgVector<vec_t,vec_size_t>& vector);

// It returns min vector's element
template <class vec_t, class vec_size_t>
vec_t  min (const mozgVector<vec_t,vec_size_t>& vector);

// It returns max vector's element
template <class vec_t, class vec_size_t>
vec_t  max (const mozgVector<vec_t,vec_size_t>& vector);

// It subtract vector2 from vector1
template <class vec_t, class vec_size_t>
void subtract (const vec_t* vector1,
	       const mozgVector<vec_t,vec_size_t>& vector2,
	       mozgVector<vec_t,vec_size_t>* result_vec);

// the same
template <class vec_t, class vec_size_t>
void subtract (const mozgVector<vec_t,vec_size_t>& vector1,
	       const vec_t* vector2,
	       mozgVector<vec_t,vec_size_t>* result_vec);

// the same
template <class vec_t, class vec_size_t>
void subtract (const mozgVector<vec_t,vec_size_t>& vector1,
	       const mozgVector<vec_t,vec_size_t>& vector2,
	       mozgVector<vec_t,vec_size_t>* result_vec);

// It computes inner(scalar) vector product of vector1 and vector2
template <class vec_t, class vec_size_t>
void multiply (const mozgVector<vec_t,vec_size_t>& vector1,
	       const mozgVector<vec_t,vec_size_t>& vector2,
	       vec_t* product);

// It puts number to all copy elements
template <class vec_t, class vec_size_t>
void copy (const vec_t number,
	   mozgVector<vec_t,vec_size_t>* copy);

// It copies orig_vector to copy
template <class vec_t, class vec_size_t>
void copy (const vec_t* orig_vector,
	   mozgVector<vec_t,vec_size_t>* copy);

// It copies "orig" vector to copy
template <class vec_t, class vec_size_t>
void copy (const mozgVector<vec_t,vec_size_t>& orig,
	   mozgVector<vec_t,vec_size_t>* copy);

// It zeroes vector
template <class vec_t, class vec_size_t>
void zero (mozgVector<vec_t,vec_size_t>* vector);

// It sorts vector
template <class vec_t, class vec_size_t>
void sort (mozgVector<vec_t,vec_size_t>* vector);

// It puts indirrect addressing vector to rank_vector
template <class vec_t, class vec_size_t>
void rank (const mozgVector<vec_t,vec_size_t>& vector,
	   mozgVector<vec_size_t,vec_size_t>* rank_vec);

// It randomizes vector by normal distribution with sigma and mean
template <class vec_t, class vec_size_t>
void randomize (mozgVector<vec_t,vec_size_t>* vector,
		const vec_t sigma,
		const vec_t mean);

//*****************************************************************************

template <class matr_t, class matr_size_t>
class mozgMatrix
{
private:

  matr_t** matrix;          // pointer to poiter to first matrix element

  matr_size_t num_rows,     // the number of matrix rows
              num_columns;  // the number of matrix columns

public:

// return the number of matrix rows
  matr_size_t rows () const
    {
      return num_rows;
    }

// return the number of matrix columns
  matr_size_t cols () const
    {
      return num_columns;
    }

// return pointer to poiter to first matrix element
  matr_t** matr () const
    {
      return matrix;
    }

// return pointer to first matrix element
  matr_t* operator [] (const matr_size_t number) const
  {
    return matrix[number];
  }

// add the extra_matr to the mozgMatrix&
  mozgMatrix& operator += (const mozgMatrix& extra_matr); 

// create matrix
  mozgMatrix (const matr_size_t number_rows,
	      const matr_size_t number_columns)
    {
      num_rows = number_rows;
      num_columns = number_columns;

      matr_t*  tmp = new matr_t[number_rows * number_columns],
	    ** mtmp;
      assert(tmp);

      matrix = new matr_t*[number_rows];
      assert(matrix);
      mtmp = matrix;
      matr_size_t num_rows = number_rows;

      while (num_rows--)
	{
	  *mtmp++ = tmp;
	  tmp += number_columns;
	}
    }

// delete matrix
  ~mozgMatrix ()
    {
      delete [] *matrix;
      delete [] matrix;
    }
};

// It is overloading "file_stream << [mozgMatrix object]"
template <class matr_t, class matr_size_t>
ofstream& operator << (ofstream& output,
		       const mozgMatrix<matr_t,matr_size_t>& matr);

// It is overloading "file_stream >> [mozgMatrix object]"
template <class matr_t, class matr_size_t>
ifstream& operator >> (ifstream& output,
		       const mozgMatrix<matr_t,matr_size_t>& matr);

// It shows matrix
template <class matr_t, class matr_size_t>
ostream& operator << (ostream& output,
		      const mozgMatrix<matr_t,matr_size_t>& matr);

// It zeroes matrix
// WARNING !!! make sure the size(=num_rows*num_columns) don't go over
// the type size !!!
template <class matr_t, class matr_size_t>
void zero (mozgMatrix<matr_t,matr_size_t>* matr);

// It copies "orig" matrix to copy
// WARNING !!! make sure the size(=num_rows*num_columns) don't go over
// the type size !!!
template <class matr_t, class matr_size_t>
void copy (const mozgMatrix<matr_t,matr_size_t>& orig,
	   mozgMatrix<matr_t,matr_size_t>* copy);

// It copies "orig" matrix to copy
// WARNING !!! make sure the size(=num_rows*num_columns) don't go over
// the type size !!!
template <class matr_t, class matr_size_t>
void copy (matr_t*  orig,
	   mozgMatrix<matr_t,matr_size_t>* copy);

// It computes sum of two matrices
// WARNING !!! make sure the size(=num_rows*num_columns) don't go over
// the type size !!!
template <class matr_t, class matr_size_t>
void sum (const mozgMatrix<matr_t,matr_size_t>& matrix1,
	  const mozgMatrix<matr_t,matr_size_t>& matrix2,
	  mozgMatrix<matr_t,matr_size_t>* sum_matr);

// It adds extra_matrix to matrix
// WARNING !!! make sure the size(=num_rows*num_columns) don't go over
// the type size !!!
template <class matr_t, class matr_size_t>
void sum (mozgMatrix<matr_t,matr_size_t>* matrix,
	  const mozgMatrix<matr_t,matr_size_t>& extra_matrix);
  
// It multiplies matrix on d
// WARNING !!! make sure the size(=num_rows*num_columns) don't go over
// the type size !!!
template <class matr_t, class matr_size_t>
void multiply (const matr_size_t d,
	       mozgMatrix<matr_t,matr_size_t>& matrix);

// It computes product of two matrices
template <class matr_t, class matr_size_t>
void multiply (const mozgMatrix<matr_t,matr_size_t>& matrix1,
	       const mozgMatrix<matr_t,matr_size_t>& matrix2,
	       mozgMatrix<matr_t,matr_size_t>* prod_matr);

// It transposes matrix
template <class matr_t, class matr_size_t>
void transpose (const mozgMatrix<matr_t,matr_size_t>& origin_matrix,
		mozgMatrix<matr_t,matr_size_t>* transpon_matrix);

// It randomizes matrix by normal distribution with sigma and mean=0
// WARNING !!! make sure the size(=num_rows*num_columns) don't go over
// the type size !!!
template <class matr_t, class matr_size_t>
void randomize (mozgMatrix<matr_t,matr_size_t>* matrix,
		const matr_t sigma,
		const matr_t mean);

//*****************************************************************************

// It computes out vector product of v1 and v2
template <class vec_t, class vec_size_t, class matr_t, class matr_size_t>
void multiply (const mozgVector<vec_t,vec_size_t>&  v1,
	       const mozgVector<vec_t,vec_size_t>&  v2,
	       mozgMatrix<matr_t,matr_size_t>* matr);

// It computes product of matrix and vector
template <class vec_t, class vec_size_t, class matr_t, class matr_size_t>
void multiply (const mozgMatrix<matr_t,matr_size_t>& matrix,
	       const mozgVector<vec_t,vec_size_t>&  vector,
	       mozgVector<vec_t,vec_size_t>*  product_vector);

// It computes product of matrix and vector for the RBF out function
template <class vec_t, class vec_size_t, class matr_t, class matr_size_t>
void multiply (const mozgMatrix<matr_t,matr_size_t>& matrix,
	       const mozgVector<vec_t,vec_size_t>&  vector,
	       const mozgMatrix<matr_t,matr_size_t>* RBFmatrix,
	       mozgVector<vec_t,vec_size_t>* product_vector);

// It computes product of vector and matrix
template <class vec_t, class vec_size_t, class matr_t, class matr_size_t>
void multiply (const mozgVector<vec_t,vec_size_t>&  vector,
	       const mozgMatrix<matr_t,matr_size_t>& matrix,
	       mozgVector<vec_t,vec_size_t>* product_vector);


// Random number generator of Park and Miller
extern mozgflt ran1(long* idum);

// It returns random regulary distributed in [argmin,argmax] number
extern mozgflt random_number(const mozgflt argmin,
			     const mozgflt argmax);

// It returns a normally distributed deviate with zero mean 
// and unit variance
extern mozgflt gasdev();

// It returns random normally distributed values with sigma dispersion
extern mozgflt gauss(const mozgflt sigma);

// It returns random normally distributed values with sigma dispersion and 
// mean value mean
extern mozgflt gauss(const mozgflt sigma,
		     const mozgflt mean);

// It returns random regular distributed signum
extern mozgflt rand_signum();

#endif // _MOZG_VECTOR_MATRIX_H_
