// MultiLayer Perceptron (MLP) simulator (methods)

#include "mozg/config.h"
#include "mozg/MLPerceptron.hh"


static char IdentString[] =
"MLPerceptron "Mozg_PACKAGE" data file "Mozg_VERSION;

static char OldIdentString[] = Mozg_PACKAGE" data file "Mozg_VERSION;

// neuron output functions (see MLPOutFunc.[hh,cc])
OUT_FUNC output_func [7] = { mozgweinet, mozgexp, mozgth, mozglin, mozgrbf,
 mozgatg, mozgrbfnet};

// neuron output functions derivatives (see MLPOutFunc.[hh,cc])
OUT_FUNC_DERIV output_func_deriv [7] = { mozgDweinet, mozgDexp, mozgDth,
 mozgDlin, mozgDrbf, mozgDatg,  mozgDrbfnet};


// compute neuron outputs from neuron nets for the current layer
void MLPerceptron::
compute_lay_outputs (const mozgVector<mozgflt,mozgint>& lay_nets,
		     const OUT_FUNC output_function,
		     const mozgVector<mozgflt,mozgint>& lay_sigmas,
		     mozgVector<mozgflt,mozgint>* lay_outputs,
		     mozgflt lay_temperature)
{
  mozgflt* nets = lay_nets.vec (),
         * outputs = lay_outputs->vec (),
         * sigmas  = lay_sigmas.vec (),
           beta = 1./lay_temperature;

  mozgint i = lay_outputs->size ();

  if (beta==1.) while (i--)
    {
      *outputs++ = (*output_function) (*nets,*sigmas);
      nets++;
      sigmas++;
    }

  else while (i--)
    {
      *outputs++ = (*output_function) ((beta * *nets),*sigmas);
      nets++;
      sigmas++;
    }
}


// compute neuron errors for the output layer
void MLPerceptron::
compute_output_errors (const mozgVector<mozgflt,mozgint>& lay_diff,
		       const mozgVector<mozgflt,mozgint>& lay_nets,
		       const mozgVector<mozgflt,mozgint>& lay_outputs,
		       const mozgVector<mozgflt,mozgint>& lay_sigmas,
		       const OUT_FUNC_DERIV output_function_deriv,
		       mozgVector<mozgflt,mozgint>* lay_errors,
		       mozgflt lay_temperature)
{
  mozgflt* diff    = lay_diff.vec (),
         * nets    = lay_nets.vec (),
         * outputs = lay_outputs.vec (),
         * errors  = lay_errors->vec (),
         * sigmas  = lay_sigmas.vec (),
           beta    = 1./lay_temperature; // I've examined access time for
                               // member-data and auto-value, the last is less.

  mozgint i = lay_errors->size();

  if (beta == 1.) while (i--)
    {
      *errors++ = *diff++ * (*output_function_deriv) (*outputs,*nets,*sigmas);
      outputs++;
      nets++;
      sigmas++;
    }

  else while (i--)
    {
      *errors++ = *diff++ * beta *
	(*output_function_deriv) (*outputs,(beta * *nets),*sigmas);
      outputs++;
      nets++;
      sigmas++;
    }
}


// compute neuron errors for the current layer
void MLPerceptron::
compute_lay_errors (const mozgVector<mozgflt,mozgint>& previously_errors,
		    const mozgMatrix<mozgflt,mozgint>& previously_weights,
		    const mozgVector<mozgflt,mozgint>& curent_nets,
		    const mozgVector<mozgflt,mozgint>& curent_outputs,
		    const mozgVector<mozgflt,mozgint>& curent_sigmas,
		    const OUT_FUNC_DERIV output_function_deriv,
		    mozgVector<mozgflt,mozgint>* curent_errors,
		    mozgflt curent_temperature)
{
  //      vec_t delta = 0.0;
  mozgflt* cur_nets    = curent_nets.vec (),
         * cur_outputs    = curent_outputs.vec (),
         * cur_errors  = curent_errors->vec (),
         * cur_sigmas  = curent_sigmas.vec (),
           beta = 1./curent_temperature;

  mozgint j = curent_outputs.size ();

  multiply (previously_errors,
	    previously_weights,
	    curent_errors);

// It multiplies current errors on the derivative output function
// and gets "layer's delta errors" named layer's errors in the mozg
  if (beta == 1.) while (j--)
    {
      *cur_errors++ *= (*output_function_deriv) (*cur_outputs,
						 *cur_nets,
						 *cur_sigmas);
      cur_outputs++;
      cur_nets++;
      cur_sigmas++;
    }
  else while (j--)
    {
      *cur_errors++ *= beta * (*output_function_deriv) (*cur_outputs,
							(beta * *cur_nets),
							*cur_sigmas);
      cur_outputs++;
      cur_nets++;
      cur_sigmas++;
    }
}


// update weights for current layer
void MLPerceptron::
adjust_lay_weights (const struct_layer& prev_layer,
		    struct_layer* cur_layer)
{
  mozgflt* lay_inputs,
         * lay_nets,
         * lay_outputs,
         * lay_errors,
           eta;
  mozgflt* lay_RBF_weights = NULL;

  mozgflt* connections,
         * cur_sigmas = NULL,
         * last_delta = NULL,
           weight_delta;

  mozgint func_number;

// chudo v per'iah :) (in russian)
  bool flag_of_alpha            = (bool) alpha,
       flag_of_sigma_Langevin   = (bool) sigma_Langevin,
       flag_of_decay            = (bool) decay,
       flag_of_NO_RBF_sigma_min = !(bool)RBF_sigma_min;

  if ((cur_layer->function_number == 4)
      || (cur_layer->function_number == 6)) {
    cur_sigmas = cur_layer->sigmas->vec ();
    lay_RBF_weights = *(cur_layer->RBFweights->matr ());
  }
  lay_nets    = cur_layer->nets->vec ();
  eta         = cur_layer->eta;
  lay_outputs = cur_layer->outputs->vec ();
  lay_errors  = cur_layer->errors->vec ();
  func_number = cur_layer->function_number;


  for (mozgint j=0; j < cur_layer->num_outputs; j++)
    {
      connections = (cur_layer ->weights->matr ())[j];
      lay_inputs  = prev_layer.outputs->vec ();
      if (flag_of_alpha) last_delta = (cur_layer->last_deltas->matr ())[j];

      for (mozgint k=0; k < prev_layer.num_outputs; k++)
	{
	  if ( !((func_number==4) || (func_number==6)) )
	    weight_delta = eta * *lay_errors * *lay_inputs;
	  else {
	    weight_delta = -1. * eta * *lay_errors * *lay_RBF_weights;
	    lay_RBF_weights++;
	  }
	  if (flag_of_alpha) weight_delta += alpha * *last_delta;

	  if (flag_of_sigma_Langevin) weight_delta += gauss (sigma_Langevin);

	  if (flag_of_decay) weight_delta -= decay * *connections;

	  *connections += weight_delta;
	  if (flag_of_alpha)
	    {
	      *last_delta = weight_delta;
	      last_delta++;
	    }
	  connections++;
	  lay_inputs++;
	}
      if ( ((cur_layer->function_number == 4)
	    ||(cur_layer->function_number == 6))
	   && (flag_of_NO_RBF_sigma_min || (*cur_sigmas > RBF_sigma_min)))
	{
	  *cur_sigmas += eta * *lay_errors * *lay_nets / (*cur_sigmas);
	  cur_sigmas++;
	  lay_nets++;
	}
      lay_outputs++;
      lay_errors++;
    }
}

// update weights for current layer by backprop without extentions
// (optimesed CPU time)
void MLPerceptron::
o_adjust_lay_weights (const struct_layer& prev_layer,
		      struct_layer* cur_layer)
{
  mozgflt* lay_inputs,
         * lay_nets,
         * lay_outputs,
         * lay_errors,
           eta;
  mozgflt* lay_RBF_weights = NULL;

  mozgflt* connections,
         * cur_sigmas = NULL,
           weight_delta;

  mozgint func_number;

// chudo v per'iah :) (in russian)
  bool flag_of_NO_RBF_sigma_min = (bool)RBF_sigma_min;

  lay_nets    = cur_layer->nets->vec ();
  eta         = cur_layer->eta;
  lay_outputs = cur_layer->outputs->vec ();
  lay_errors  = cur_layer->errors->vec ();
  func_number = cur_layer->function_number;

  if ((func_number == 4) || (func_number == 6)) {
    cur_sigmas = cur_layer->sigmas->vec ();
    lay_RBF_weights = *(cur_layer->RBFweights->matr ());
  }

  for (mozgint j=0; j < cur_layer->num_outputs; j++)
    {
      connections = (cur_layer ->weights->matr ())[j];
      lay_inputs  = prev_layer.outputs->vec ();

      for (mozgint k=0; k < prev_layer.num_outputs; k++)
	{
	  if ( !((func_number==4) || (func_number==6)))
	    weight_delta = eta * *lay_errors * *lay_inputs;
	  else {
	    weight_delta = -1. * eta * *lay_errors * *lay_RBF_weights;
	    lay_RBF_weights++;
	  }
	  *connections += weight_delta;
	  connections++;
	  lay_inputs++;
	}
      if ( ((func_number == 4) || (func_number == 6))
	  && (flag_of_NO_RBF_sigma_min || (*cur_sigmas > RBF_sigma_min)))
	{
	  *cur_sigmas += eta * *lay_errors * *lay_nets / (*cur_sigmas);
	  cur_sigmas++;
	  lay_nets++;
	}
      lay_outputs++;
      lay_errors++;
    }
}

// propagate signal through *UPPER layer
void  MLPerceptron::
propLayer (const struct_layer& LOWER,
	   struct_layer* UPPER)
{
    if ( !((UPPER->function_number == 4)
	   || (UPPER->function_number == 6)))
      {
	multiply (*UPPER->weights,
		  *LOWER.outputs,
		  UPPER->nets);

	compute_lay_outputs (*UPPER->nets,
			     UPPER->lay_output_func,
			     *UPPER->sigmas, // meaningless
			     UPPER->outputs,
			     UPPER->temperature);
      }
    else
      {
	multiply (*UPPER->weights,
		  *LOWER.outputs,
		  UPPER->RBFweights,
		  UPPER->nets);

	compute_lay_outputs (*UPPER->nets,
			     UPPER->lay_output_func,
			     *UPPER->sigmas,
			     UPPER->outputs,
			     UPPER->temperature);
      }
}

// propagate input vector forward through network to the outputs and save
// outputs in the outvec
void MLPerceptron::
propInputs (mozgflt* invec,
	    mozgflt* outvec)
{
  putInputs (invec);

  for (mozgint i=1; i < num_layers; i++)
    {
      propLayer (layer[i-1],
		 &layer[i]);
      if ( flag_of_bias_term && (i != (num_layers-1)) )
	layer[i].outputs->vec () [layer[i].num_outputs-1] = 1.;
    }
  getOutputs (outvec);
}

// propagate vectors of the invecarr forward through network to the outputs
// and save outputs for each vector in outvecarr
void MLPerceptron::
propInputs (mozgflt** invecarr,
	    mozgflt** outvecarr,
	    mozgint ssize)
{
  for (mozgint vecnum=0; vecnum<ssize; vecnum++) {
    putInputs (invecarr[vecnum]);

    for (mozgint i=1; i < num_layers; i++)
      {
	propLayer (layer[i-1],
		   &layer[i]);
	if ( flag_of_bias_term && (i != (num_layers-1)) )
	  layer[i].outputs->vec () [layer[i].num_outputs-1] = 1.;
      }
    getOutputs (outvecarr[vecnum]);
  }
}

// test network
mozgflt MLPerceptron::
testNet (mozgflt** tinp,
	 mozgflt** tout,
	 mozgint tssize)
{
  mozgflt testerror=0.;

  for (mozgint vecnum=0; vecnum<tssize; vecnum++) {
    putInputs (tinp[vecnum]);

    for (mozgint i=1; i < num_layers; i++)
      {
	propLayer (layer[i-1],
		   &layer[i]);
	if ( flag_of_bias_term && (i != (num_layers-1)) )
	  layer[i].outputs->vec () [layer[i].num_outputs-1] = 1.;
      }
    compOutputErrors (tout[vecnum],false);

    testerror += error_over_outputs;
  }
  return (testerror/tssize);
}

// compute network output errors and place square of the error vector module in
// the error_over_outputs
void MLPerceptron::
compOutputErrors (mozgflt* target,
		  bool outerrflag)
{
  struct_layer& lay = layer[num_layers - 1];

  subtract (target,
	    *lay.outputs,
	    diff);

  error_over_outputs = 0.;

// place square of a error vector module in the error_over_outputs
  multiply (*diff,
	    *diff,
	    &error_over_outputs);

  switch (switch_of_energy_function) {

  case 2: copy((mozgflt)1.,temp_vector);

    subtract(*temp_vector,*lay.outputs,temp_vector);

    *temp_vector *= *lay.outputs;

    *diff /= *temp_vector;break;

  case 3: *diff /= (1 - error_over_outputs);break;

  case 4: copy(target,temp_vector);     // temp_vector_i=t_i

    *temp_vector /= *lay.outputs;break; // temp_vector_i=t_i/o_i
  }
  if (outerrflag) 

    switch (switch_of_energy_function) {

    case 1:case 2:case 3: compute_output_errors (*diff,
						 *lay.nets,
						 *lay.outputs,
						 *lay.sigmas,
						 layer[num_layers - 1]
						   .lay_output_func_deriv,
						 lay.errors,
						 lay.temperature);break;

    case 4:        compute_output_errors (*temp_vector,
					  *lay.nets,
					  *lay.outputs,
					  *lay.sigmas,
					  layer[num_layers - 1]
					    .lay_output_func_deriv,
					  lay.errors,
					  lay.temperature);
    }
}

// propagate out errors back through the network
void MLPerceptron::
backpropErrors ()
{
  struct_layer* cur_layer,
              * prev_layer;

  mozgMatrix<mozgflt,mozgint>* prev_weights;

  for (mozgint i = num_layers - 1; i > 1; i--)
    {
      prev_layer  = &layer[i];
      cur_layer   = &layer[i - 1];

      prev_weights = ((prev_layer->function_number==4)
		      ||(prev_layer->function_number==6)) ?
	  prev_layer->RBFweights
	: prev_layer->weights;

      compute_lay_errors (*prev_layer->errors,
			  *prev_weights,
			  *cur_layer->nets,
			  *cur_layer->outputs,
			  *cur_layer->sigmas,
			  cur_layer->lay_output_func_deriv,
			  cur_layer->errors,
			  cur_layer->temperature);
    }
}

// adjust network weights
void MLPerceptron::
adjustWeights ()
{
  struct_layer* cur_layer,
              * prev_layer;

  for (mozgint i=1; i < num_layers; i++)
    {
      prev_layer = &layer[i - 1];
      cur_layer  = &layer[i];

      if (flag_of_optimisation)
	o_adjust_lay_weights (*prev_layer,
			      cur_layer);

      else adjust_lay_weights(*prev_layer,
			      cur_layer);
    }
  if (sigma_Langevin) sigma_Langevin *= sigma_Langevin_decrease;
}

// learn net
void MLPerceptron::
learnNet (mozgflt* linp,
	  mozgflt* lout)
{
  putInputs (linp);

  for (mozgint i=1; i < num_layers; i++)
    {
      propLayer (layer[i-1],
		 &layer[i]);
      if ( flag_of_bias_term && (i != (num_layers-1)) )
	layer[i].outputs->vec () [layer[i].num_outputs-1] = 1.;
    }
  compOutputErrors (lout,true);

  backpropErrors ();

  adjustWeights ();
}

// learn net
void MLPerceptron::
learnNet (mozgflt** linp,
	  mozgflt** lout,
	  mozgint lssize,
	  mozgint order,
	  mozgint epnum,
	  mozgint pepnum)
{
  mozgint vecnum,
          random_vecnum = 0;

  for (mozgint tprint=1,i=1; i<=epnum; tprint++, i++, overall_epnum++) {

    mozgflt lerr = 0.;

    for (vecnum=0; vecnum < lssize; vecnum++) {

      if (!order) {

	random_vecnum = (mozgint)random_number(0.,(mozgflt)lssize);

	putInputs (linp[random_vecnum]);
      }
      else  putInputs (linp[vecnum]);

      for (mozgint j=1; j < num_layers; j++)
	{
	  propLayer (layer[j-1],
		     &layer[j]);
	  if ( flag_of_bias_term && (j != (num_layers-1)) )
	    layer[j].outputs->vec () [layer[j].num_outputs-1] = 1.;
	}
      if (!order) compOutputErrors (lout[random_vecnum],true);

      else compOutputErrors (lout[vecnum],true);

      lerr += error_over_outputs;

      backpropErrors ();

      adjustWeights ();
    }
    if (tprint==pepnum) {

      tprint = 0;
// finish (1) or continue (0) learning...
      if (answerMessage((lerr/lssize),(mozgflt)0.)) i = epnum + 1;
    }
  }
}

// learn and test net
void MLPerceptron::
learnNet (mozgflt** linp,
	  mozgflt** lout,
	  mozgint lssize,
	  mozgint order,
	  mozgint epnum,
	  mozgint pepnum,
	  mozgflt** tinp,
	  mozgflt** tout,
	  mozgint tssize)
{
  mozgint vecnum,
          random_vecnum = 0;

  for (mozgint tprint=1,i=1; i<=epnum; tprint++, i++, overall_epnum++) {

    mozgflt lerr = 0.;

    for (vecnum=0; vecnum < lssize; vecnum++) {

      if (!order) {

	random_vecnum = (mozgint)random_number(0.,(mozgflt)lssize);

	putInputs (linp[random_vecnum]);
      }
      else putInputs (linp[vecnum]);

      for (mozgint j=1; j < num_layers; j++)
	{
	  propLayer (layer[j-1],
		     &layer[j]);
	  if ( flag_of_bias_term && (j != (num_layers-1)) )
	    layer[j].outputs->vec () [layer[j].num_outputs-1] = 1.;
	}
      if (!order) compOutputErrors (lout[random_vecnum],true);

      else compOutputErrors (lout[vecnum],true);

      lerr += error_over_outputs;

      backpropErrors ();

      adjustWeights ();
    }
    if (tprint==pepnum) {

      tprint = 0;
      mozgflt tsqer = 0.;

      for (vecnum=0; vecnum < tssize; vecnum++) {

	putInputs (tinp[vecnum]);

	compOutputErrors (tout[vecnum],false);

	tsqer += error_over_outputs;
      }
// finish (1) or continue (0) learning...
      if (answerMessage((lerr/lssize),(tsqer/tssize))) i = epnum + 1;
    }
  }
}

// create network, use "vanilla" backprop in learning, if RBF isn't used,
// RBF_sigma_min doesn't influence on computations
MLPerceptron::
MLPerceptron(mozgint   number_of_layers,
	     mozgint*  num_outputs_layers,
	     mozgint*  func_number,
	     mozgflt   weight_dispersion,
	     bool bias_term_flag)
{
  struct_layer* lay;

  num_layers = number_of_layers;
  layer     = new struct_layer[num_layers];
  assert (layer);

  flag_of_last_deltas = false;
  flag_of_bias_term = bias_term_flag;
  flag_of_temp_vector = false;
  is_call_aM = false;

  overall_epnum = 1;

// Default values:
  for (mozgint i=1; i<num_layers; i++) {

    layer[i].eta = 0.3 / num_outputs_layers[i];
    layer[i].temperature = 1.;
  }
  RBF_sigma_min             = 0.;
  switch_of_energy_function = 1;
  alpha                     = 0.;
  sigma_Langevin            = 0.;
  sigma_Langevin_decrease   = 0.;
  decay                     = 0.;
  flag_of_optimisation      = true;
  error_threshold           = 0.;
  rate_coeff = 1.;
  temperature_coeff = 1.;

// allocate network out_errors
  diff = new mozgVector<mozgflt,mozgint> (num_outputs_layers[num_layers-1]);

// make 1,... network layers, randomize weights[j][k]...
  for (mozgint i=0; i < num_layers; i++ )
    {
      lay = &(layer[i]);

      lay->num_outputs = num_outputs_layers[i];

      lay->outputs = new mozgVector<mozgflt,mozgint>(lay->num_outputs);

      if (i>0) {
	lay->function_number = func_number[i];
	lay->lay_output_func = output_func[func_number[i]];
	lay->lay_output_func_deriv = output_func_deriv[func_number[i]];

	lay->nets = new mozgVector<mozgflt,mozgint>(num_outputs_layers[i]);

	lay->weights = new mozgMatrix<mozgflt,mozgint> (num_outputs_layers[i],
						  num_outputs_layers[i-1]);

	lay->errors = new mozgVector<mozgflt,mozgint> (num_outputs_layers[i]);

	lay->sigmas = new mozgVector<mozgflt,mozgint> (num_outputs_layers[i]);


	if ( (lay->function_number == 4)
	     ||(lay->function_number == 6))
	  {
	    lay->RBFweights = 
	      new mozgMatrix<mozgflt,mozgint> (num_outputs_layers[i],
					       num_outputs_layers[i-1]);
	  }
	else {
	  lay->RBFweights = new mozgMatrix<mozgflt,mozgint> (1,1);
	}
	randomize  (layer[i].weights,
		    // I've believed first layer has max units
		    weight_dispersion,
		    (mozgflt)0.);

	randomize  (layer[i].sigmas,
		    (mozgflt)1e-3,
		    (mozgflt)1.);

      }
    }
}


// delete network
MLPerceptron::
~MLPerceptron ()
{
  diff->~mozgVector ();
  layer[0].outputs->~mozgVector ();

  for (mozgint i=1; i < num_layers; i++)
    {
      layer[i].outputs->~mozgVector ();
      layer[i].nets->~mozgVector ();

      if (i > 0)
	{
	  layer[i].weights->~mozgMatrix ();

	  layer[i].sigmas->~mozgVector ();

	  layer[i].RBFweights->~mozgMatrix ();

	  layer[i].errors->~mozgVector ();

	  if (alpha) layer[i].last_deltas->~mozgMatrix ();
	}
    }
  delete [] layer;

  if ((switch_of_energy_function==2)
      || (switch_of_energy_function==4)) temp_vector->~mozgVector ();
}

// It is overloading "file_stream << [MLPerceptron object]"
ofstream& operator << (ofstream& output,
		       const MLPerceptron& net)
{
  mozgint i = 0;

  if (output.good ()) {
    output << IdentString << "\n";

    if (net.flag_of_bias_term) output << "bias term\n";
    else output << "\n";

    output << "the number of layers: " << net.num_layers;

    output << "\nlayers neurons number:";

    for (i=0; i<net.num_layers; i++) {
      output  << ' ' << net.layer[i].num_outputs;
    }
    output << "\nlayers function number:";
    //    i = net.num_layers;
    for (i=0; i<net.num_layers; i++) {
      output << ' ' << net.layer[i].function_number;
    }
    output << endl;

    for (i=1; i<net.num_layers; i++) {

      output << "\nlayer [" << i  << "]:\n";

      output << (*net.layer[i].weights);

      if ((net.layer[i].function_number==4)
	  ||(net.layer[i].function_number==6))
	output << (*net.layer[i].sigmas) << endl;
    }
  }
  else {
    cerr << "\nI'm congratulating you."
	 << "You have some trouble with net saving ;)\n";
// ? exception handler call...
  }
  return output;
}

// It is overloading "file_stream >> [MLPerceptron object]"
ifstream& operator >> (ifstream& input,
		       MLPerceptron** p_p_net)
{
  mozgint i;
  mozgint  num_layers,
         * num_neurons,
         * function_num;

  bool bias_flag = false;

  char temp [50];

  if (!input.good ()) {
    cerr << "\nFailed to open \"";
// ? exception handler call...
  }
  else {
    input.getline (temp, 50, '\n');
// compare first 12 symbols of the temp and IdentString
// && temp and OldIdentString
    if (strncmp (temp,IdentString,27)
	&& strncmp (temp, OldIdentString,12)) {
      cerr << "Unsupported file version, or garbage in file.\n";
// ? exception handler call...
    }
    else {
//      input.ignore(1);
      input.getline (temp, 50, '\n');
      if (strcmp (temp, "bias term")) bias_flag = false;
      else if (strcmp (temp, "")) bias_flag = true;
           else cerr << "Reading error: no \"\""
		     << " (or \"bias term\") line.\n";

      input.getline (temp, 50, ':');
      input >> num_layers;
      if (strcmp (temp, "the number of layers")) {
	cerr << "Reading error: no \"the number of layers: \" line.\n";
// ? exception handler call...
      }
      else {
	num_neurons = new mozgint[num_layers];
	function_num = new mozgint[num_layers];

	input.getline (temp, 50, ':');
	for (i=0; i<num_layers; i++) {
	  input >> num_neurons[i] ;
	}
	input.getline (temp, 50, ':');
	for (i=0; i<num_layers; i++) {
	  input >> function_num[i];
	}
	*p_p_net = new MLPerceptron (num_layers,
				     num_neurons,
				     function_num,
				     (mozgflt)0.,
				     bias_flag);

	for (i=1; i<num_layers; i++) {
	  input.ignore(2);
	  input.getline (temp, 50, '\n');
	  input >> (*(**p_p_net).layer[i].weights);
	  if (!input.good ()) {
	    cerr << "Premature EOF on data file\n";
// ? exception handler call...
	  }
	  if ((function_num[i]==4)
	      || (function_num[i]==6)) {
	    input >> (*(**p_p_net).layer[i].sigmas);
	  }
	}
      }
    }
  }
  return input;
}

// show network on display
ostream& operator << (ostream& output,
		      const MLPerceptron& net)
{
  mozgint i = 0;

  if (output.good ()) {

    output << "number of layers: " << net.num_layers;

    output << "\nneurons on layer numbers:";
    i = net.num_layers;
    for (i=0; i<net.num_layers; i++) {
      output  << ' ' << net.layer[i].num_outputs;
    }

    output << "\nnumbers of layer function:";
    i = net.num_layers;
    for (i=0; i<net.num_layers; i++) {
      output << ' ' << net.layer[i].function_number;
    }
    output << endl;

    if (net.layer[1].eta) {
      output << "eta   = " << net.layer[1].eta << endl;
      output << "alpha = " << net.alpha << endl;
      output << "decay = " << net.decay << endl;
    }
    for (i=1; i<net.num_layers; i++) {

      output << "\nlayer [" << i  << "]:\n";

      output << (*net.layer[i].weights);

      if (net.layer[i].function_number==4)
	output << (*net.layer[i].sigmas) << endl;
    }
  }
  else {
    cerr << "\nI'm congratulating you."
	 << "You have some trouble with net showing ;)\n";
// ? exception handler call...
  }
  return output;
}
